#! /usr/bin/env python

# Genetic
# Copyright (C) 2001 Jean-Baptiste LAMY
#
# This program is free software. See README or LICENSE for the license terms.

"""Genetic demo 2

This demo shows how to minimize function with constraints : we just consider
that any organism that doesn't fit the constraints cannot live. This correspond
to a None phenotype.

Here, we minimize cos(x) where 0.0 < x < 6.0 (for info, the minimum
is PI = 3.14...).
"""

import random, string, math
from genetic import organism, lifecycle


# Set the function to minimize. We choose the cosinus function, between 0.0 and 6.0 -- the
# minimum is PI !

def FUNC(x):
  if x < 0.0 or x > 6.0:
    # Organism with "None" phenotype cannot live !
    return None
  return math.cos(x)


class FUNCMinimizer(organism.Organism):
  def compute_phenotypes(self):
    self.func_value, self.func_args = organism.recessive_phenotype(self.genotype, FUNC)
    if self.func_value is None:
      # This organism is not viable.
      self.canlive = 0
      
  def __cmp__(self, other):
    return cmp(self.func_value, other.func_value)
  
  def __repr__(self):
    return "Minimum: FUNC%s = %s\n" % (self.func_args, self.func_value) + organism.Organism.__repr__(self)


# Create the initial set of organisms.
#
# As usual in Python, __<something>__ genes are some "magic" value that
# influence the chromosom replication or stability, see
# organism.Chromosom.__doc__ for a list of them.
#
# What is really MAGIC is that the are genes... so they can mutate ! Yes, the
# probability of mutation can change ! This is what make the organisms
# adaptate very well.
#
# Those magic genes are set per-chromosom, so if you have genes with different
# scales or units, it can be a good idea to put them on different chromosom
# (so they can have different magic values).
#
# Notice that, if __mutation__ and/or __mutampl__ fall down, the organisms
# won't continue their evolution, and they will refuse to go on, even if they
# are not at the "end" of the problem ! So yes, they are free to "fix" them
# where they are, and stop evoluting ! This is also a kind of adaptation... ;-)

organismA = FUNCMinimizer([
  organism.Chromosom(__mutation__ = 0.5, __mutampl__ = 4.0, x = 5.0),
  ])

organismB = FUNCMinimizer([
  organism.Chromosom(__mutation__ = 1.0, __mutampl__ = 2.0, x = 1.0),
  ])

organisms = [organismA, organismB]


# Define some parameters

# True for elitism (= always keep the best organism)
ELITISM = 1

# Number of generation to run
NB_GENERATION = 8

# Number of children each generation made
NB_CHILDREN  = 75

# Number of organisms kept per generation (other children doesn't survive)
NB_ORGANISMS = 10


# Ready !
#
# This will make live a lot of organism, and finally print the last and "best"
# one -- the one whose phenotype minimizes FUNC !

lifecycle.run(organisms, ELITISM, NB_GENERATION, NB_CHILDREN, NB_ORGANISMS)

# You should approach the PI value.
#
# Obviously, mathematicians will say that the cos function already include the
# PI value, so we do nothing... let them speak and let our organisms live !
