#! /usr/bin/env python

# Genetic
# Copyright (C) 2001 Jean-Baptiste LAMY
#
# This program is free software. See README or LICENSE for the license terms.

import random, string, math, Tkinter, thread
from time import sleep
from genetic import organism, lifecycle

SIZE = 250
OBSTACLE_HEIGHT = 100 # You may try some harder challanger, e.g. 150. Or no obstacle, e.g. 0.

class JumperOrganism(organism.Organism):
  def __init__(self, genotype):
    organism.Organism.__init__(self, genotype)
    self.reset()
    
  def reset(self):
    self.x = 0.0
    self.y = float(SIZE)
    self.jumptime = None
    self.arrived = 0
    
  def compute_phenotypes(self):
    self.ax = organism.perchromosom_dominancy_gene(self.genotype, "ax")
    self.bx = organism.perchromosom_dominancy_gene(self.genotype, "bx")
    self.cx = organism.perchromosom_dominancy_gene(self.genotype, "cx")
    self.ay = organism.perchromosom_dominancy_gene(self.genotype, "ay")
    self.by = organism.perchromosom_dominancy_gene(self.genotype, "by")
    self.cy = organism.perchromosom_dominancy_gene(self.genotype, "cy")
    if self.ax is None or self.bx is None or self.cx is None or self.ay is None or self.by is None or self.cy is None:
      self.canlive = 0
    
  def __cmp__(self, other):
    # Say how we compare 2 organisms. Here, the lower the circle radius is the better the organism is.
    return cmp(self.jumptime, other.jumptime)
  
  def __repr__(self):
    return "Time to arrive : %s\n" % self.jumptime + organism.Organism.__repr__(self)
    
  def play(self, time):
    dx = self.ax * self.x + self.bx * self.y + self.cx
    dy = self.ay * self.x + self.by * self.y + self.cy
    
    length = math.sqrt(dx ** 2 + dy ** 2) / 2.0
    
    newx = self.x + dx / length
    newy = self.y + dy / length
    
    if canpass(newx, newy): self.x, self.y = newx, newy
    else:
      if   canpass(newx, self.y): self.x = newx
      elif canpass(self.x, newy): self.y = newy
      
    if arrived(self.x, self.y):
      self.arrived  = 1
      self.jumptime = time
      

def randomjumper():
  return JumperOrganism([
    organism.Chromosom(ax = rand(), bx = rand(), cx = rand(), __mutampl__ = 0.1, __mutation__ = 0.8, __break__ = 0.0),
    organism.Chromosom(ay = rand(), by = rand(), cy = rand(), __mutampl__ = 0.1, __mutation__ = 0.8, __break__ = 0.0),
    ])

def rand(): return (random.random() - 0.5) / 30.0

def canpass(x, y):
  return 0 < x < SIZE and 0 < y < SIZE and not (100 < x < 150 and y > (SIZE - OBSTACLE_HEIGHT))

def arrived(x, y): return x > 220 and y > 150


class Graph(Tkinter.Canvas):
  def __init__(self, master):
    Tkinter.Canvas.__init__(self, master, bg = "white", width = SIZE, height = SIZE)
    
    self.obstacle  = self.create_rectangle(100, SIZE - OBSTACLE_HEIGHT, 150, 250, fill = "black");
    self.organisms = [randomjumper() for i in range(20)]
    
    thread.start_new_thread(self.run, ())
    
  def run(self):
    while 1:
      tags = []
      
      for organism in self.organisms:
        organism.tag = self.create_oval(organism.x - 3, organism.y - 3, organism.x + 3, organism.y + 3, outline = "blue")
        tags.append(organism.tag)
       
      for time in range(500): # Let give them 500 moves !
        nooneleft = 1
        
        for organism in self.organisms:
          if not organism.arrived:
            oldx, oldy = organism.x, organism.y
            organism.play(time)
            self.move(organism.tag, organism.x - oldx, organism.y - oldy)
            nooneleft = 0
            
        if nooneleft: break
        sleep(0.01)
        
      # Remove the looser (jumptime is None)
      self.organisms = filter(lambda organism: not organism.jumptime is None, self.organisms)
      
      # Sort them (the first are the better)
      self.organisms.sort()

      # Dump info
      if self.organisms: print self.organisms[0]
      print "%s organisms arrived at destination !" % len(self.organisms)
      print
      
      # Creates 20 new organisms, whose parents are the five first winners of the previous generation.
      self.organisms = self.organisms[:5]
      while len(self.organisms) < 2: self.organisms.append(randomjumper()) # We need at least 2 organisms !
      self.organisms = lifecycle.make_loves(self.organisms, 20)
      
      self.delete(*tags)
      

class GraphFrame(Tkinter.Tk):
  def __init__(self):
    Tkinter.Tk.__init__(self, className = "genetic jump !")
    Tkinter.Label(self, text = "Organisms starts on the lower left corner; they must arrive at the lower right corner. The rectangle is an obstacle they must jump !\nAfter some generations, organisms will learn how to jump !", wraplength = SIZE + 30).pack()
    self.graph = Graph(self)
    self.graph.pack()
    

graphframe = GraphFrame()
Tkinter.mainloop()
