#! /usr/bin/env python

# Genetic
# Copyright (C) 2001 Jean-Baptiste LAMY
#
# This program is free software. See README or LICENSE for the license terms.

from __future__ import nested_scopes

from genetic import organism, lifecycle
import Tkinter, operator, math


SIZE = 250.0
STEP =  10.0

class Point:
  nextid = 0
  def __init__(self, x, y, value = 0.0):
    self.x     = x
    self.y     = y
    self.value = value
    self.id     = self.nextid
    self.nextid = self.nextid + 1
    
  def distanceto(self, x, y):
    return math.sqrt((self.x - x) ** 2 + (self.y - y) ** 2)
    
  def __repr__(self): return "<Point %s, %s value %s>" % (self.x, self.y, self.value)
  
points = []

def value2color(value):
  if   value < -1.0: value = -1.0
  elif value >  1.0: value =  1.0
  return "gray" + str(int((value + 1.0) * 50.0))


def value_at(point_factors, point): # returns the theorical value for a point.
  return value_at_xy(point_factors, point.x, point.y)

def value_at_xy(point_factors, x, y): # returns the theorical value at x, y.
  return reduce(operator.add, map(lambda point: math.exp(-point.distanceto(x, y) / 100.0) * point_factors[point.id] * point.value, points))
  
def square(point_factors):
  return reduce(operator.add, map(lambda point: (value_at(point_factors, point) - point.value) ** 2, points), 0.0)


class PointsSeparator(organism.Organism):
  characteristics = [
    organism.Characteristic("square", square, organism.RECESSIVE_PHENOTYPE)
    ]
  
  def __cmp__(self, other):
    # Say how we compare 2 organisms. Here, the lower the square is, the better the organism is.
    return cmp(self.square, other.square)


class Graph(Tkinter.Canvas):
  def __init__(self, master):
    Tkinter.Canvas.__init__(self, master, bg = "gray50", width = SIZE, height = SIZE)
    self.bind("<Button-1>", self.leftClick )
    self.bind("<Button-3>", self.rightClick)
    self.rects = []
    self.best  = None
    self.reset()
    
  def reset(self):
    organismA = PointsSeparator([
      organism.Chromosom(point_factors = [0.9] * len(points), __mutampl__ = 0.5),
      ])
    
    organismB = PointsSeparator([
      organism.Chromosom(point_factors = [1.1] * len(points), __mutampl__ = 1.0),
      ])
    
    self.organisms = [organismA, organismB]
    
  def leftClick  (self, event): self.addPoint(Point(event.x, event.y, -1.0))
  def rightClick (self, event): self.addPoint(Point(event.x, event.y,  1.0))
  
  def addPoint(self, point):
    points.append(point)
    self.create_rectangle(point.x - 3, point.y - 3, point.x + 4, point.y + 4, fill = "gray50", outline = "")
    if point.value > 0.0: self.create_text(point.x, point.y, text = "+", fill = "white")
    else:                 self.create_text(point.x, point.y, text = "+", fill = "black")
    # Phenotypes have changed, because there is one point more !
    self.reset()
    
  def generation(self, nb = 1):
    self.organisms = lifecycle.run(self.organisms, 1, nb, 40, 15)
    
  def update(self):
    best = self.organisms[0]
    if not best is self.best: # If it is still the same, it's useless to update !
      self.best = self
      
      point_factors, = best.square_args
      #point_factors = [1.0]
      
      self.delete(self.rects)
      DEMI_STEP = STEP / 2.0
      
      for i in range(0, SIZE, STEP):
        for j in range(0, SIZE, STEP):
          theorical_value = value_at_xy(point_factors, i + DEMI_STEP, j + DEMI_STEP)
          self.rects.append(self.create_rectangle(i, j, i + STEP, j + STEP, tags = "rect", outline = "", fill = value2color(theorical_value)))
          
      self.tag_lower("rect")
      

class GraphFrame(Tkinter.Tk):
  def __init__(self):
    Tkinter.Tk.__init__(self, className = "genetic separator")
    Tkinter.Label(self, text = "Left-click below to add a black point, and right-click to add a white one, then run 1 or 10 generations !\nI will genetically separe black and white points !", wraplength = SIZE + 30).pack()
    self.graph = Graph(self)
    self.graph.pack()
    Tkinter.Button(self, text = "1   generation" , command = self.generation1  ).pack()
    Tkinter.Button(self, text = "10  generations", command = self.generation10 ).pack()
    
  def generation1(self, event = None):
    self.graph.generation(1)
    self.graph.update()
        
  def generation10(self, event = None):
    self.graph.generation(10)
    self.graph.update()
    

graphframe = GraphFrame()
Tkinter.mainloop()
