#! /usr/bin/env python

# Genetic
# Copyright (C) 2001 Jean-Baptiste LAMY
#
# This program is free software. See README or LICENSE for the license terms.

import random, string, math, Tkinter, thread
from genetic import organism, lifecycle

SIZE = 250
OBSTACLE_Y1 = 250
OBSTACLE_Y2 = 125
OBSTACLE_HEIGHT1 = 125
OBSTACLE_HEIGHT2 = 125

OBSTACLE_Y      = OBSTACLE_Y1
OBSTACLE_HEIGHT = OBSTACLE_HEIGHT1

class JumperOrganism(organism.Organism):
  def __init__(self, genotype):
    organism.Organism.__init__(self, genotype)
    self.reset()
    
  def reset(self):
    self.x = 0.0
    self.y = float(SIZE) / 2.0
    self.jumptime = None
    self.arrived = 0
    
  def compute_phenotypes(self):
    self.ax = organism.perchromosom_dominancy_gene(self.genotype, "ax")
    self.bx = organism.perchromosom_dominancy_gene(self.genotype, "bx")
    self.cx = organism.perchromosom_dominancy_gene(self.genotype, "cx")
    self.ay = organism.perchromosom_dominancy_gene(self.genotype, "ay")
    self.by = organism.perchromosom_dominancy_gene(self.genotype, "by")
    self.cy = organism.perchromosom_dominancy_gene(self.genotype, "cy")
    if self.ax is None or self.bx is None or self.cx is None or self.ay is None or self.by is None or self.cy is None:
      self.canlive = 0
    
  def play(self, time):
    dx = self.ax * self.x + self.bx * self.y + self.cx
    dy = self.ay * self.x + self.by * self.y + self.cy
    
    length = math.sqrt(dx ** 2 + dy ** 2) / 2.0
    
    newx = self.x + dx / length
    newy = self.y + dy / length
    
    if canpass(newx, newy): self.x, self.y = newx, newy
    else:
      if   canpass(newx, self.y): self.x = newx
      elif canpass(self.x, newy): self.y = newy
      
    if arrived(self.x, self.y):
      self.arrived  = 1
      self.jumptime = time
      

def randomjumper():
  return JumperOrganism([
    organism.Chromosom(ax = rand(), __mutampl__ = 0.1, __mutation__ = 0.8, __break__ = 0.0),
    organism.Chromosom(bx = rand(), __mutampl__ = 0.1, __mutation__ = 0.8, __break__ = 0.0),
    organism.Chromosom(cx = rand(), __mutampl__ = 0.1, __mutation__ = 0.8, __break__ = 0.0),
    organism.Chromosom(ay = rand(), __mutampl__ = 0.1, __mutation__ = 0.8, __break__ = 0.0),
    organism.Chromosom(by = rand(), __mutampl__ = 0.1, __mutation__ = 0.8, __break__ = 0.0),
    organism.Chromosom(cy = rand(), __mutampl__ = 0.1, __mutation__ = 0.8, __break__ = 0.0),
    ])

def rand(): return (random.random() - 0.5) / 30.0

def canpass(x, y):
  return 0 < x < SIZE and 0 < y < SIZE and not (100 < x < 150 and (OBSTACLE_Y - OBSTACLE_HEIGHT) < y < OBSTACLE_Y)

def arrived(x, y): return x > 220


class Graph(Tkinter.Canvas):
  def __init__(self, master):
    Tkinter.Canvas.__init__(self, master, bg = "white", width = SIZE, height = SIZE)
    self.obstacle  = None
    self.organisms = [randomjumper() for i in range(20)]
    
    thread.start_new_thread(self.run, ())
    
  def set_obstacle(self, y, height):
    global OBSTACLE_Y, OBSTACLE_HEIGHT
    OBSTACLE_Y      = y
    OBSTACLE_HEIGHT = height
    if self.obstacle: self.delete(self.obstacle)
    self.obstacle = self.create_rectangle(100, OBSTACLE_Y - OBSTACLE_HEIGHT, 150, OBSTACLE_Y, fill = "black");
    
  def run(self):
    while 1:
      # Move the obstacle.
      if OBSTACLE_Y == OBSTACLE_Y1: self.set_obstacle(OBSTACLE_Y2, OBSTACLE_HEIGHT2)
      else: self.set_obstacle(OBSTACLE_Y1, OBSTACLE_HEIGHT1)
      
      tags = []
      
      for organism in self.organisms:
        organism.tag = self.create_oval(organism.x - 3, organism.y - 3, organism.x + 3, organism.y + 3, outline = "blue")
        tags.append(organism.tag)
       
      for time in range(500): # Let give them 500 moves !
        nooneleft = 1
        
        for organism in self.organisms:
          if not organism.arrived:
            oldx, oldy = organism.x, organism.y
            organism.play(time)
            self.move(organism.tag, organism.x - oldx, organism.y - oldy)
            nooneleft = 0
            
        if nooneleft: break
        
      # Remove the looser (jumptime is None)
      self.organisms = filter(lambda organism: not organism.jumptime is None, self.organisms)
      
      # Dump info
      if self.organisms: print self.organisms[0]
      print "%s organisms arrived at destination !" % len(self.organisms)
      print
      
      # Creates 20 new organisms, whose parents are the organism's survivors.
      if len(self.organisms) == 0: self.organisms.append(randomjumper())
      self.organisms = lifecycle.make_loves(self.organisms, 20)
      
      self.delete(*tags)
    

class GraphFrame(Tkinter.Tk):
  def __init__(self):
    Tkinter.Tk.__init__(self, className = "genetic wall !")
    Tkinter.Label(self, text = "Organisms start on the left; they must arrive on right. The rectangle is a wall; each turn, the wall moves !\nAfter some (lots of) generations, 2 species of organisms will distinguish !", wraplength = SIZE + 30).pack()
    self.graph = Graph(self)
    self.graph.pack()
    

graphframe = GraphFrame()
Tkinter.mainloop()
