# Genetic
# Copyright (C) 2001 Jean-Baptiste LAMY
#
# This program is free software. See README or LICENSE for the license terms.

"""genetic.organism -- Organism and Chromosom class.

An Organism is described by his 'genotype' and his 'phenotype'. The organism
class should be inherited to create your own organisms; see its __doc__.

A 'genotype' is considered to be a list of pairs of chromosoms : [(chromosom_0_A, chromosom_0_B), (chromosom_1_A, chromosom_1_B), ...]
Some chromosoms can be 'None', for incomplete pairs.
The genotype is inherited from the parents (A child take one chromosom of each pair of each parent).

A 'phenotype' is a 2-values tuple; the first value is the result of the
function that you want to minimize for the organism, the second is the
sequence of args given to this function .

The phenotype is computed from the genotype, with the help of a phenotype
function. See recessive_phenotype, dominant_phenotype and
perchromosom_dominancy_phenotype for such functions.
"""

#from __future__ import nested_scopes
import random, types, copy, operator, math


# OLD DEPRECATED STUFF !

RECESSIVE_PHENOTYPE = 1
DOMINANT_PHENOTYPE = 2
PERCHROMOSOM_DOMINANCY_PHENOTYPE = 3

class Characteristic:
  """A characteristic of an organism : e.g. eyes color, ...
The characteritic's phenotype is computed from the genotype.
This class should be inherited/extended to create different, and :
A characteristic id defined by :
 - The "func" function that computes its value. "func" takes genes values
   and return the phenotype. The args of "func" should have the same names
   that genes.
 - The "phenotype" function that choose between the multiple values for a
   genes the one used by "func" : there's typically 2 values for the same gene,
   but only one can be used by "func" !
   Common phenotype functions are provided in this class, use the following
   constant in the constructor : RECESSIVE_PHENOTYPE (the default),
   DOMINANT_PHENOTYPE and PERCHROMOSOME_DOMINANCY_PHENOTYPE.
"""
  def __init__(self, name, func, phenotype = RECESSIVE_PHENOTYPE, funcargs = None):
    """Characteristic(name, func, phenotype = RECESSIVE_PHENOTYPE, funcargs = None) -> Characteristic -- Create a new characteristic.
funcargs is the list of the name of the genes that correspond to "func" argument.
If not provided, funcargs is guessed from func's arguments names.
"""
    self.name = name
    self.func = func
    
    if funcargs is None:
      if not type(func) is types.FunctionType: func = func.im_func
      self.funcargs = func.func_code.co_varnames[: func.func_code.co_argcount]
    else: self.funcargs = funcargs

    if   phenotype is RECESSIVE_PHENOTYPE: self.phenotype = self.recessive_phenotype
    elif phenotype is DOMINANT_PHENOTYPE : self.phenotype = self.dominant_phenotype
    elif phenotype is PERCHROMOSOM_DOMINANCY_PHENOTYPE:
      self.phenotype = self.perchromosom_dominancy_phenotype
    else: self.phenotype = phenotype
    
  def perchromosom_dominancy_phenotype(self, genotype):
    """perchromosom_dominancy_phenotype(genotype) -> phenotype -- Gets the dominant phenotype from the given list, this func use per-chromosom dominancy (__dominancy__ gene).
Notice that this phenotype computation method is cheaper in time that dominant or recessive method."""
    # 1 - Collects all the values for each gene, in a dict : {"gene1" : (dominancy, value), ...}
    attrs = {}
    for pair in genotype:
      for chromosom in pair:
        if chromosom is None: continue
        for gene, value in chromosom.__dict__.items():
          attr = attrs.get(gene, None)
          if (not attr is None) and (attr[0] >= chromosom.__dominancy__):
            # The previous found version of this gene is on a chromosom more dominant that the current one... skip !
            continue
          
          attrs[gene] = (chromosom.__dominancy__, value)
  
    # 2 - Gets all the attrs needed by func in a list : [(dominancy1, value1), ...]
    attrs_for_func = map(attrs.get, self.funcargs)
    
    # 3 - Gets all the args in a list : [value1, value2, ...]
    try: args = [attr[1] for attr in attrs_for_func]
    except TypeError:
      # A needed gene is not present ! (= attr is None)
      return (None, None)
    
    # 4 - Computes the phenotype
    return (self.func(*args), args)
  
  def recessive_phenotype(self, genotype):
    """recessive_phenotype(genotype) -> phenotype -- Gets the better phenotype from the given list. This roughly corresponds to a recessive disease."""
    # Compute all the phenotype, and choose the one with the minimal value -- the best one
    def chooser(phenotypes):
      phenotypes = filter(lambda phenotype: not phenotype[0] is None, phenotypes)
      if phenotypes: return min(phenotypes)
      return (None, None)
      
    return self._compute_all_phenotypes(genotype, chooser)
  
  phenotype = recessive_phenotype
  
  def dominant_phenotype(self, genotype):
    """dominant_phenotype(genotype) -> phenotype -- Gets the worst phenotype from the given list. This roughly corresponds to a dominant disease."""
    # Compute all the phenotype, and choose the one with the maximal value -- the worst one
    def chooser(phenotypes):
      for phenotype in phenotypes:
        if phenotype[0] is None: return phenotype
      return max(phenotypes)
    
    return self._compute_all_phenotypes(genotype, chooser)
  
  def _compute_all_phenotypes(self, genotype, chooser):
    """_compute_all_phenotypes(genotype, chooser) -> phenotype -- An utility func that compute ALL the possible phenotype, and let chooser choose the right one.
genotype is a list of chromosoms pair, and chooser a function that take a list of phenotypes as argument, and should return the right one."""
    # 1 - Collects all the values for each gene, in a dict : {"gene1":[value1, value2, ...], ...}
    attrs = {}
    for pair in genotype:
      for chromosom in pair:
        if chromosom is None: continue
        for gene, value in chromosom.__dict__.items():
          attr = attrs.get(gene, None)
          if attr is None:
            attr = []
            attrs[gene] = attr
          attr.append(value)
    
    # Computes the phenotype from the genotype :
    # 2 - Gets all the args in a list : [[x1, x2], [y1, y2], ...]
    attrs_for_func = map(attrs.get, self.funcargs)
    
    # 3 - Gets all the combinations of these args : [[x1, y1], [x1, y2], ..., [x2, y1], [x2, y2], ...]
    try: allcombinations = combinations(*attrs_for_func)
    except EmptyListError:
      # All the needed genes are not present... this organism cannot live !
      return (None, None)
    
    # 4 - Computes all the possible phenotypes : [(result, [x1, y1]), ...]
    phenotypes = [(self.func(*combination), combination) for combination in allcombinations]
    
    # 5 - Choose the right one.
    return chooser(phenotypes)
    
# END OF OLD DEPRECATED STUFF.



def perchromosom_dominancy_gene(genotype, gene):
  """perchromosom_dominancy_phenotype(genotype, gene) -> gene_value
For the given gene, returned the value of the most dominant copy (There are
multiple (commonly 2) value for the same gene in the genotype).

This phenotype-computation function use per-chromosom dominancy (__dominancy__
gene) : it chooses the gene that is on the chromosom with the highest
__dominancy__ value.

genotype is a sequence of chromosoms pairs.
gene is the name of the desired gene.

This function should be used instead of phenotype functions, when you don't
have a func to minimize and you are doing more sophisticated selection, like
in artificial life, ... -- see demo_jump.py for an example.
"""
  value     = None
  dominancy = -1.0e99
  for pair in genotype:
    for chromosom in pair:
      if (not chromosom is None) and chromosom.__dominancy__ > dominancy:
        try:
          value = getattr(chromosom, gene)
          dominancy = chromosom.__dominancy__
        except (AttributeError, NameError, KeyError): pass
        
  return value

def perchromosom_dominancy_phenotype(genotype, func, funcargnames = None):
  """perchromosom_dominancy_phenotype(genotype, func, funcargnames) -> phenotype
Computes the phenotype of a genotype, with the given func. This function
doesn't compute the value of the phenotype (hey, THIS is YOUR job !), but
it chooses the genes for you (remember, there are typically 2 values for each
gene as there are 2 chromosoms in each pair; so you need to choose one value
in the 2. This function does it).

This phenotype-computation function use per-chromosom dominancy (__dominancy__
gene) : it chooses the gene that is on the chromosom with the highest
__dominancy__.

genotype is a sequence of chromosoms pairs.
func is the function used to compute the phenotype value; it should take at
least one argument.
funcargnames are the sequence of the genes's names that correspond to func
argument; if not given, the args's name of func are used.

The returned phenotype is a tuple, whose first value is the value returned by
func and whose second value is the sequence of arguments given to func to
obtain this value.

Notice that this phenotype computation method is cheaper in time that dominant
or recessive method, but not as good..."""
  # 1 - Gets the name of func arguments
  if funcargnames is None: funcargnames = func.func_code.co_varnames[:func.func_code.co_argcount]
  
  # 2 - Collects all the values for each gene, in a dict : {"gene1" : (dominancy, value), ...}
  attrs = {}
  for pair in genotype:
    for chromosom in pair:
      if chromosom is None: continue
      for gene, value in chromosom.__dict__.items():
        attr = attrs.get(gene, None)
        if (not attr is None) and (attr[0] >= chromosom.__dominancy__):
          # The previous value found for this gene is on a more dominant chromosom... skip !
          continue
        
        attrs[gene] = (chromosom.__dominancy__, value)
        
  # 3 - Gets all the attrs needed by func in a list : [(dominancy1, value1), ...]
  attrs_for_func = map(attrs.get, funcargnames)
  
  # 4 - Gets all the args in a list : [value1, value2, ...]
  try: args = [attr[1] for attr in attrs_for_func]
  except TypeError:
    # A needed gene is not present ! (= attr is None)
    return None, None
  
  # 5 - Computes the phenotype
  return func(*args), args

def recessive_phenotype(genotype, func, funcargnames = None):
  """recessive_phenotype(genotype, func, funcargnames = None) -> phenotype
Computes the phenotype of a genotype, with the given func. This function
doesn't compute the value of the phenotype (hey, THIS is YOUR job !), but
it chooses the genes for you (remember, there are typically 2 values for each
gene as there are 2 chromosoms in each pair; so you need to choose one value
in the 2. This function does it).

This phenotype-computation function chooses the best combinations of genes :
the one for which func give the smallest value.
This roughly corresponds to a recessive disease.

genotype is a sequence of chromosoms pairs.
func is the function used to compute the phenotype value; it should take at
least one argument.
funcargnames are the sequence of the genes's names that correspond to func
argument; if not given, the args's name of func are used.

The returned phenotype is a tuple, whose first value is the value returned by
func and whose second value is the sequence of arguments given to func to
obtain this value."""
  # 1 - Computes all the possible phenotypes, and removes the bad one.
  phenotypes = filter(lambda phenotype: not phenotype[0] is None,
                      _compute_all_phenotypes(genotype, func, funcargnames)
                      )
  
  # 2 - Returns the smallest (=the best) phenotype.
  if phenotypes: return min(phenotypes) # a phenotype is a tuple
  return None, None # The non-viable phenotype

def dominant_phenotype(genotype, func, funcargnames = None):
  """dominant_phenotype(genotype, func, funcargnames = None) -> phenotype
Computes the phenotype of a genotype, with the given func. This function
doesn't compute the value of the phenotype (hey, THIS is YOUR job !), but
it chooses the genes for you (remember, there are typically 2 values for each
gene as there are 2 chromosoms in each pair; so you need to choose one value
in the 2. This function does it).

This phenotype-computation function chooses the worst combinations of genes :
the one for which func give the highest value.
This roughly corresponds to a dominant disease.

genotype is a sequence of chromosoms pairs.
func is the function used to compute the phenotype value; it should take at
least one argument.
funcargnames are the sequence of the genes's names that correspond to func
argument; if not given, the args's name of func are used.

The returned phenotype is a tuple, whose first value is the value returned by
func and whose second value is the sequence of arguments given to func to
obtain this value."""
  # 1 - Computes all the possible phenotypes, and removes the bad one.
  phenotypes = filter(lambda phenotype: not phenotype[0] is None,
                      _compute_all_phenotypes(genotype, func, funcargnames)
                      )
  
  # 2 - Returns the smallest (=the best) phenotype.
  if phenotypes: return max(phenotypes) # a phenotype is a tuple
  return None, None # The non-viable phenotype

def _compute_all_phenotypes(genotype, func, funcargnames = None):
  """_compute_all_phenotypes(genotype, func, funcargnames = None) -> list of phenotypes
An utilitary function that compute ALL the possible phenotypes.
"""
  # 1 - Gets the name of func arguments
  if funcargnames is None: funcargnames = func.func_code.co_varnames[:func.func_code.co_argcount]
  
  # 2 - Collects all the values for each gene, in a dict : {"geneX":[x1, x2, ...], "geneY":[y1, y2, ...], ...}
  attrs = {}
  for pair in genotype:
    for chromosom in pair:
      if chromosom is None: continue
      for gene, value in chromosom.__dict__.items():
        attr = attrs.get(gene, None)
        if attr is None:
          attr = []
          attrs[gene] = attr
        attr.append(value)
  
  # 3 - Gets all the args in a list : [[x1, x2], [y1, y2], ...] where x is the first argument of func, and y the second
  attrs_for_func = map(attrs.get, funcargnames)
  
  # 4 - Gets all the combinations of these args : [[x1, y1], [x1, y2], ..., [x2, y1], [x2, y2], ...]
  try: allcombinations = combinations(*attrs_for_func)
  except EmptyListError:
    # All the needed genes are not present... this organism cannot live !
    return ()
  
  # 5 - Computes all the possible phenotypes : [(result, [x1, y1]), ...]
  return [(func(*combination), combination) for combination in allcombinations]
  

class Organism:
  """The Organism class. An organism has :
 - a genotype = a list of (possibly incomplete) pair of chromosoms,
 - a phenotype, computed by the compute_phenotypes method; this method is
   called by __init__.

You must inherit this class, and :
 - provide a compute_phenotypes() method : the phenotype functions of this
   module are here to help you; see demo.
 - provide a comparison method, like __cmp__(other) : to compare 2 organisms.
   Better organisms should compare "smaller" (and so are the first in a sorted
   list).
 - provide a print method, like __repr__() : you probably want to print
   interesting information on your organisms. Organism.__repr__ just prints
   the genotype, you may want to call it here.
"""
  characteristics = []
  
  def __init__(self, genotype):
    """Organism(genotype) -> Organism -- Creates a new Organism with the given genotype.
genotype can be either a list of pairs of chromosoms, or a list of chromosoms
for homozygote organisms (= both chromosoms of each pair are the same)."""
    # 1 - Save data
    if len(genotype) > 0 and isinstance(genotype[0], Chromosom):
      # genotype is not a list of pairs of chromosoms but a list of chromosom
      # => this organism is homozygote
      genotype = map(None, genotype, genotype)
    self.genotype = genotype
    
    # 2 - Remove useless pair of chromosoms from the genotype -- this may not be really fair... ??
    for pair in self.genotype[:]:
      if (pair[0] is None or pair[0].useless()) and (pair[1] is None or pair[1].useless()): self.genotype.remove(pair)
    
    # 3 - Compute the phenotypes
    self.canlive = 1
    self.compute_phenotypes()
    
  def compute_phenotypes(self):
    # Compute the phenotypes for each characteristic
    for characteristic in self.characteristics:
      value, args = characteristic.phenotype(self.genotype)
      setattr(self, characteristic.name, value)
      setattr(self, characteristic.name + "_args", args) # The args that give this result -- usefull ! It's often what we want to know.
      
      # If a phenotype returns None, there is no available phenotype, so the organism cannot live.
      if value is None: self.canlive = 0
      
  def __repr__(self):
    if self.canlive:
      repr = reduce(operator.add,
        ["%s%s : %s%s\n" % (characteristic.name, characteristic.funcargs, getattr(self, characteristic.name), getattr(self, characteristic.name + "_args")) for characteristic in self.characteristics],
        ""
        )
      
    else: repr = "phenotype : DEAD (non-viable)\n"
      
    i = 0
    for pair in self.genotype:
      if pair[0] is None: repr = repr + "genotype, chromosome %s A : None\n" % i
      else:               repr = repr + "genotype, chromosome %s A :\n%s\n" % (i, `pair[0]`)
      if pair[1] is None: repr = repr + "genotype, chromosome %s B : None\n" % i
      else:               repr = repr + "genotype, chromosome %s B :\n%s\n" % (i, `pair[1]`)
      i = i + 1
      
    return repr
  
  def __eq__(self, other): return self.genotype == other.genotype
  
  def givetochild(self):
    """Organism.givetochild() -> list
Returns the list of chromosom that this organism give to a child.
It is the half of a full genome."""
    gift = []
    for pair in self.genotype:
      chromosom = self.givechromosomtochild(pair)
      if isinstance(chromosom, Chromosom): gift.append(chromosom)
      else: gift.extend(chromosom)
    return gift
  
  def givechromosomtochild(self, pair):
    """Organism.givechromosomtochild(pair) -> chromosom
Chooses, between a pair of chromosom, the one to give to a child."""
    # Check if there is only one chromosom.
    if pair[0] is None: return pair[1] or []
    if pair[1] is None: return pair[0]
    
    if random.random() < (pair[0].__crossover__ + pair[1].__crossover__) / 2.0:
      # Crossing over !
      chromosom = pair[0].crossover(pair[1])
    else:
      # Choose a random chromosom
      if random.random() < 0.5: chromosom = pair[0]
      else:                     chromosom = pair[1]
      
    # Checks for mutation
    return chromosom.mutate()
  

class Mutable:
  """Mixin class for any mutable object that is not a chromosom.
This allows to use Mutable object as gene's value, instead of float.
See genetic.mutablelist.MutableOrderList for an example.
"""
  def mutate(self, chromosom):
    """Mutable.mutate(chromosom) -> new value -- Called to check mutation in this mutable object.
The chromosom param is the chromosom whose magic genes should be used to
control the mutation (e.g. __mutation__). The implementation can still used its
own value if they are more accurate.
The returned value should be the mutable object itself if it hasn't been
changed.
"""
      

class Chromosom(Mutable):
  """The Chromosom class, that can be used in the genotype.
Support many magic genes, and automatically mutate ALL its attributes when
necessary. E.g. to add a gene called eyecolor with the RGB value [0.0, 0.5, 0.8],
just do :

>>> chromosom.eyecolor = [0.0, 0.5, 0.8]

The velue in the list will mutate as apropriate. You can give the gene to the
constructor :

>>> chromosom = Chromosom(eyecolor = [0.0, 0.5, 0.8])

the DEFAULT_GENES attribute of the Chromosom class (not instance) contains
some genes (typically magic genes like __mutation__) that are automatically
added to each Chromosom. Child class may want to override this dictionnary.

Here is the magic genes supported by Chromosom :
__mutation__ is the pobability for a gene to mutate while reproduction,
ranging from 0.0 to 1.0.

__mutampl__ is the amplitude of the mutation. A gene that mutate is modified
by + or - a random number ranging from 0.0 to __mutampl__. Notice that
__mutampl__ doesn't apply to magic genes.

__mutsign__ is the sign of the mutation. If __mutsign__ is 0.0 (default) or
if 0.5 < __mutsign__ < 2.0, the mutation can be + or - (it can increase or
decrease the gene value). If __mutsign__ < 0.5, the mutation can only
decrease the value; if > 2.0, it can only increase.

__crossover__ is the probability that a cross-over occurs -- a cross-over
will mix 2 chromosom.

__deletion__ is the probability to delete a random gene -- organism doesn't
like that...

__loss__ is the probability for the chromosom to be lost -- organism doesn't
like it too !

__break__ is the probability for the chromosom to be broken in 2 parts. If
you use recessive phenotype function, organisms often like this one, as it
allows them to have more genetic stuff -- so more chance to survive ! But
chromosom breaking can cost A LOT of computation time ! So it may be a good
idea to disable it (__break__ = 0.0).

__dominancy__ is the dominancy of the chromosom. This is only use by the
phenotype function organism.perchromosom_dominancy_phenotype, which choose
genes from the most dominant chromosom.
"""
  DEFAULT_GENES = {
    "__dominancy__" :  1.0 ,
    "__mutation__"  :  0.5 ,
    "__mutampl__"   : 50.0 ,
    "__mutsign__"   :  0.0 ,
    "__crossover__" :  0.2 ,
    "__break__"     :  0.01,
    "__deletion__"  :  0.01,
    "__loss__"      :  0.01,
    }
  
  def __init__(self, **data):
    """Chromosom("gene name" = value) -> new Chromosom."""
    self.__dict__.update(self.DEFAULT_GENES)
    self.__dict__.update(data)
    
  def useless(self):
    """Chromosom.useless() -> boolean -- Check if the Chromosom is useless.
A useless chromosom is a chromosom that has ONLY magic genes (=__(something)__).
"""
    for gene in self.__dict__.keys():
      if not gene.startswith("__"): return 0
    return 1
  
  def crossover(self, other):
    """Chromosom.crossover(other_chromosom) -> Chromosom -- Perform a cross-over between this chromosom and an other one.
"""
    dict = {}
    genes1 = self .__dict__.items()
    genes2 = other.__dict__.items()
    crossat = int(random.random() * len(genes1))
    
    i = 0
    while i < crossat:
      dict[genes1[i][0]] = genes1[i][1]
      i = i + 1
    while i < len(genes2):
      dict[genes2[i][0]] = genes2[i][1]
      i = i + 1
    
    return Chromosom(**dict)
  
  def mutate(self, chromosom = None):
    """Chromosom.mutate() -> new value -- Called to check mutation in this chromosom.
The returned value is the chromosom itself if it hasn't been changed.
For compatibility with the Mutable class, a (useless) chromosom param can be passed.
"""
    # Checks if the chromosom is lost
    if random.random() < self.__loss__: return []
    
    newdict = self.mutate_object(self.__dict__)
    if newdict: mutated = self.__class__(**newdict)
    else:       mutated = self
      
    # Checks if a gene is deleted on the chromosom
    if random.random() < self.__deletion__:
      if mutated is self: mutated = self.__class__(**self.__dict__)
      
      deletablegenes = filter(lambda gene: not gene.startswith("__"), mutated.__dict__.keys())
      if len(deletablegenes) > 0:
        delattr(mutated, random.choice(deletablegenes))
    
    # Checks if the chromosom breaks in 2 parts
    if random.random() < self.__break__:
      genes = mutated.__dict__.items()
      breakat = int(random.random() * (len(genes) + 1))
      
      genes1, genes2 = {}, {}
      for i in range(breakat): genes1[genes[i][0]] = genes[i][1]
      for i in range(breakat, len(genes)): genes2[genes[i][0]] = genes[i][1]
      return self.__class__(**genes1), self.__class__(**genes2)
    
    return mutated
  
  def mutate_object(self, object, name = ""):
    """Chromosom.mutate_object(object, name = "") -> new value
Mutate the object parameter with the magic genes of this Mutable object.
object should be a float value, a Mutable object or a list or a dict whose
values can be passed as argument to mutate_object.
The returned value should be the object param itself if it hasn't been
changed.
"""
    objtype = type(object)
    if objtype is types.FloatType:
      if random.random() > self.__mutation__: return object
      
      if name.startswith("__"):
        # The gene is a "magic" gene. For magic gene, the amplitude of mutation is the gene's value itself.
        return object * math.exp(random.gauss(0.0, 1.0))
        #return object * (.5 + random.random() * 1.0) # this old func give more chance to decrease than to inscrease !
      else:
        # Else, use the default mutation sign and amplitude (= the value of the "__mutsign__" and "__mutampl__" gene)
        mutsign = self.__mutsign__
        
        if mutsign == 0.0 or (mutsign > 0.5 and mutsign < 2.0):
          # No sign for the mutation
          return object + random.gauss(0.0, 1.0) * self.__mutampl__
          #return object + (random.random() - 0.5) * 2.0 * self.__mutampl__
        else:
          # The mutation has a sign : if sign < 0.5, mutation can only decrease the gene's value. Else increase it.
          if mutsign < 0.5: return object - random.random() * self.__mutampl__
          else:             return object + random.random() * self.__mutampl__
          
    elif isinstance(object, Mutable): return object.mutate(self)
    
    elif objtype is types.DictType:
      mutated = object
      for gene, value in object.items():
        newvalue = self.mutate_object(value, gene)
        if not newvalue is value:
          if mutated is object: mutated = object.copy()
          mutated[gene] = newvalue
      return mutated
    
    elif objtype is types.ListType:
      mutated = map(self.mutate_object, object)
      if mutated != object: return mutated
      return object
    
    else: return object # Non mutable => do not mutate !
    
  def __repr__(self):
    genes = self.__dict__.items()
    genes.sort()
    
    return reduce(operator.add,
      map(
        lambda (gene, value): "    %s\t : %s\n" % (gene, value),
        filter(lambda (gene, value): value != 0.0 or not gene.startswith("__"), genes)
        )
      ).expandtabs(32)[:-1] # remove the last \n.
  

def make_love(organismA, organismB):
  """make_love(organismA, organismB) --> Organism
Returns the child of organismA and organismB !"""
  return organismA.__class__(map(None, organismA.givetochild(), organismB.givetochild()))


# Tools functions :

EmptyListError = "EmptyListError"

def combinations(*lists):
  """combinations([x1, x2,...], [y1, y2, ...], ...) -> [[x1, y1], [x1, y2], ..., [x2, y1], [x2, y2], ...]
Gets all the possible combinations, from the given lists of args.
The first value of each tuple is taken from the first list argument, the seond from the second, ..."""
  if not lists[0]: raise EmptyListError
  if len(lists) == 1: return zip(lists[0])
  
  r = []
  subcombinations = combinations(*lists[1:])
  for first in lists[0]:
    r.extend([[first] + list(combination) for combination in subcombinations])
    
  return r

def mean(*floats):
  """mean(a, b, c, ...) -> float
Returns the mean of a list of number."""
  return reduce(operator.add, floats) / len(floats)

